/******************************************************************************\
 * Technische Universitaet Darmstadt, Institut fuer Nachrichtentechnik
 * Copyright (c) 2001
 *
 * Author(s):
 *	Volker Fischer
 *
 * Description:
 *	
 *
 ******************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
\******************************************************************************/

#include <qlistview.h>
#include <qpixmap.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qcombobox.h>
#include <qurloperator.h>
#include <qnetworkprotocol.h>
#include <qdir.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qlayout.h>
#include <qftp.h>
#include <qthread.h>
#include <qwt_counter.h>
#include <qaction.h>
#include <qwhatsthis.h>
#include <qlabel.h>
#include <qwt_thermo.h>

#ifdef _WIN32
# include "../../Windows/moc/StationsDlgbase.h"
#else
# include "moc/StationsDlgbase.h"
#endif
#include "../DrmReceiver.h"
#include "../Vector.h"

#ifdef HAVE_LIBHAMLIB
# include <hamlib/rig.h>
#endif


extern CDRMReceiver	DRMReceiver;


/* Definitions ****************************************************************/
/* Define the timer interval of updating the list view */
#define GUI_TIMER_LIST_VIEW_STAT		30000 /* ms (30 seconds) */

/* Define the timer interval of updating the UTC label */
#define GUI_TIMER_UTC_TIME_LABEL		1000 /* ms (1 second) */

/* Define the timer interval of updating s-meter */
#define GUI_TIMER_S_METER				300 /* ms (0.3 seconds) */

/* s-meter thermo parameters */
#define S_METER_THERMO_MIN				((_REAL) -60.0) /* dB */
#define S_METER_THERMO_MAX				((_REAL) 60.0) /* dB */
#define S_METER_THERMO_ALARM			((_REAL) 0.0) /* dB */

/* Location of the newest DRM schedule ini-file "DRMSchedule.ini". This file
   is generated by Klaus Schneider which is the administrator of
   www.drm-dx.de */
#define DRM_SCHEDULE_UPDATE_FILE		"ftp://216.92.35.131/DRMSchedule.ini"

/* File handle type */
#ifdef _WIN32
# define FILE_HANDLE					HANDLE
#else
# define FILE_HANDLE					int
#endif

#ifdef HAVE_LIBHAMLIB
/* Config string for com-port selection is different in Windows and Linux */
# ifdef _WIN32
#  define HAMLIB_CONF_COM1				"rig_pathname=COM1"
#  define HAMLIB_CONF_COM2				"rig_pathname=COM2"
#  define HAMLIB_CONF_COM3				"rig_pathname=COM3"
# else
#  define HAMLIB_CONF_COM1				"rig_pathname=/dev/ttyS0"
#  define HAMLIB_CONF_COM2				"rig_pathname=/dev/ttyS1"
#  define HAMLIB_CONF_COM3				"rig_pathname=/dev/ttyUSB0"
# endif
#endif


/* Classes ********************************************************************/
class CStationsItem
{
public:
	int GetStartTimeNum() {return iStartHour * 100 + iStartMinute;}
	int GetStopTimeNum() {return iStopHour * 100 + iStopMinute;}
	void SetStartTimeNum(const int iStartTime)
	{
		/* Recover hours and minutes */
		iStartHour = iStartTime / 100;
		iStartMinute = iStartTime - iStartHour * 100;
	}
	void SetStopTimeNum(const int iStopTime)
	{
		/* Recover hours and minutes */
		iStopHour = iStopTime / 100;
		iStopMinute = iStopTime - iStopHour * 100;
	}

	int		iStartHour;
	int		iStartMinute;
	int		iStopHour;
	int		iStopMinute;
	int		iFreq;
	int		iDays;
	string	strName;
	string	strTarget;
	string	strLanguage;
	string	strSite;
	string	strCountry;
	_REAL	rPower;
};


class CDRMSchedule
{
public:
	CDRMSchedule();
	virtual ~CDRMSchedule() {}

	void ReadStatTabFromFile(const string strFileName);

	int GetStationNumber() {return StationsTable.Size();}
	CStationsItem& GetItem(int const iPos) {return StationsTable[iPos];}
	_BOOLEAN IsActive(int const iPos);

protected:
	CVector<CStationsItem> StationsTable;
};


class MyListViewItem : public QListViewItem
{
public:
	MyListViewItem(QListView* parent, QString s1, QString s2 = QString::null,
		QString s3 = QString::null, QString s4 = QString::null,
		QString s5 = QString::null, QString s6 = QString::null,
		QString s7 = QString::null, QString s8 = QString::null) :
		QListViewItem(parent, s1, s2, s3, s4, s5, s6, s7, s8) {}

	/* Custom "key()" function for correct sorting behaviour */
	virtual QString key(int column, bool ascending) const;
};


class StationsDlg : public CStationsDlgBase
{
	Q_OBJECT

public:
	StationsDlg(QWidget* parent = 0, const char* name = 0, bool modal = FALSE,
		WFlags f = 0);
	virtual ~StationsDlg();

protected:
	enum ERemotecntr {RC_NOREMCNTR, RC_WINRADIO, RC_AOR7030, RC_ELEKTOR304,
		RC_JRC_NRD535, RC_TT_RX320D};
	enum ECOMNumber {CN_COM1, CN_COM2, CN_COM3};

	void			SetFrequency(const int iFreqkHz);

	_BOOLEAN		SetFrequencyWinradio(const int iFreqkHz);
	_BOOLEAN		SetFrequencyAOR7030(const ECOMNumber eCOMNumber, const int iFreqkHz);
	_BOOLEAN		SetFrequencyNRD535(const ECOMNumber eCOMNumber, const int iFreqkHz);
	_BOOLEAN		SetFrequencyRX320D(const ECOMNumber eCOMNumber, const int iFreqkHz);

	enum EOutWire {OW_TXD, OW_DTR, OW_RTS};
	_BOOLEAN		SetFrequencyElektor304(const ECOMNumber eCOMNumber, const int iFreqkHz);
	void			OutputElektor304(FILE_HANDLE hCom, const uint32_t iData);
	void			SetOutStateElektor304(FILE_HANDLE hCom, EOutWire eOutWire, _BINARY biState);

	void			SetStationsView();
    virtual void	showEvent(QShowEvent* pEvent);
	void			AddWhatsThisHelp();
	void			SetUTCTimeLabel();
	void			EnableSMeter(_BOOLEAN bStatus);

	CDRMSchedule				DRMSchedule;
	QPixmap						BitmCubeGreen;
	QPixmap						BitmCubeYellow;
	QPixmap						BitmCubeRed;
	QTimer						TimerList;
	QTimer						TimerUTCLabel;
	QTimer						TimerSMeter;
	_BOOLEAN					bShowAll;
	QUrlOperator				UrlUpdateSchedule;
	QPopupMenu*					pViewMenu;
	CVector<MyListViewItem*>	vecpListItems;

	/* Remote selection */
	QPopupMenu*			pRemoteMenu;
	ERemotecntr			eWhichRemoteControl;

	/* Com port selection */
	QAction*			pacMenuCOM1;
	QAction*			pacMenuCOM2;
	QAction*			pacMenuCOM3;
	QActionGroup*		agCOMPortSel;
	ECOMNumber			eComNumber;

#ifdef HAVE_LIBHAMLIB
	class CSpecDRMRig
	{
	public:
		CSpecDRMRig() : iModelID(0), strDRMSet(""), iFreqOffs(0) {}
		CSpecDRMRig(const CSpecDRMRig& nSpec) : iModelID(nSpec.iModelID),
			strDRMSet(nSpec.strDRMSet),  iFreqOffs(nSpec.iFreqOffs) {}
		CSpecDRMRig(rig_model_t newID, string sSet, int iNFrOff) :
		  iModelID(newID), strDRMSet(sSet), iFreqOffs(iNFrOff) {}

		rig_model_t		iModelID; /* Model ID for hamlib */
		string			strDRMSet; /* Special DRM settings string */
		int				iFreqOffs; /* Frequency offset */
	};
	CVector<CSpecDRMRig>	vecSpecDRMRigs;

	QPopupMenu*				pRemoteMenuOther;
	CVector<rig_model_t>	veciModelID;

	class SDrRigCaps
	{
	public:
		SDrRigCaps() : iModelID(0), strManufacturer(""), strModelName(""),
			eRigStatus(RIG_STATUS_ALPHA) {}
		SDrRigCaps(rig_model_t tNID, QString strNMan, QString strNModN,
			rig_status_e eNSt) : iModelID(tNID),  strManufacturer(strNMan),
			strModelName(strNModN), eRigStatus(eNSt) {}

		inline SDrRigCaps& operator=(const SDrRigCaps& cNew)
			{iModelID = cNew.iModelID; strManufacturer = cNew.strManufacturer;
			strModelName = cNew.strModelName; eRigStatus = cNew.eRigStatus;
			return *this;}

		rig_model_t		iModelID;
		QString			strManufacturer;
		QString			strModelName;
		rig_status_e	eRigStatus;
	};
	CVector<SDrRigCaps>	veccapsHamlibModels;
	void SortHamlibModelList(CVector<SDrRigCaps>& veccapsHamlibModels);

	rig_model_t			iCurSelModelID;

	static int			PrintHamlibModelList(const struct rig_caps* caps, void* data);
	_BOOLEAN			CheckForSpecDRMFE(const rig_model_t iID, int& iIndex);
	_BOOLEAN			SetFrequencyHamlib(const int iFreqkHz);
	void				InitHamlib(const rig_model_t newModID);

	RIG*				pRig;
#endif

public slots:
	void OnRemoteMenu(int iID);
	void OnTimerList();
	void OnTimerUTCLabel() {SetUTCTimeLabel();}
	void OnTimerSMeter();
	void OnListItemClicked(QListViewItem* item);
	void OnUrlFinished(QNetworkOperation* pNetwOp);
	void OnShowStationsMenu(int iID);
	void OnComPortMenu(QAction* action);
	void OnGetUpdate();
	void OnFreqCntNewValue(double dVal);
};
